/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#ifndef DEFS_H
#define DEFS_H

#include <iostream>
#include <string>
#include <utility>
#include <vector>

namespace std {
  template<typename T>
  struct hash<vector<T>>
  {
    size_t operator()(const vector<T>& vec) const {
      size_t h{0};
      const auto elemhasher = hash<T>{};
      for (const auto& elem: vec) {
        h ^= elemhasher(elem) + 0x9e3779b9 + (h << 6) + (h >> 2);
      }
      return h;
    }
  };
  template<typename T1, typename T2>
  struct hash<pair<T1,T2>>
  {
    size_t operator()(const pair<T1,T2>& p) const {
      size_t h{0};
      h ^= hash<T1>{}(p.first) + 0x9e3779b9 + (h << 6) + (h >> 2);
      h ^= hash<T2>{}(p.second) + 0x9e3779b9 + (h << 6) + (h >> 2);
      return h;
    }
  };
}


namespace pyred {

  // stream insertion operators

template<typename T1, typename T2>
std::ostream& operator<<(std::ostream&, const std::pair<T1,T2>&);

template<typename T>
inline std::ostream& operator<<(std::ostream& out, const std::vector<T>& vec) {
  out << "[";
  for (auto it = vec.cbegin(); it != vec.cend(); ++it) {
    if (it != vec.cbegin()) out << ",";
    out << *it;
  }
  out << "]";
  return out;
}

template<typename T1, typename T2>
inline std::ostream& operator<<(std::ostream& out, const std::pair<T1,T2>& p) {
  out << "<" << p.first << "," << p.second << ">";
  return out;
}

  
class init_error: public std::exception {
private:
  std::string msg;
public:
  inline init_error(const std::string& s): msg(s) {}
  virtual inline const char* what() const noexcept {
    return msg.c_str();
  }
};

class parser_error: public std::exception {
public:
  std::string token;
  inline parser_error(const std::string& s): token(s) {}
  virtual inline const char* what() const noexcept {
    return (std::string("invalid token in coefficient: ") + token).c_str();
  }
};

class zero_modular_division: public std::exception {
public:
  inline zero_modular_division() {}
  virtual inline const char* what() const noexcept {
    return "Modular arithmetic lead to a spurious zero division.";
  }
};

class input_error: public std::exception {
private:
  std::string msg;
public:
  inline input_error(const std::string& s): msg(s) {}
  virtual inline const char* what() const noexcept {
    return msg.c_str();
  }
};


// type aliases for the integral class component

class IntegralPyred;

using pow_type = int;
using weight_type = uint64_t;
using vs_equation = std::vector<std::pair<IntegralPyred,std::string>>;
using ws_equation = std::vector<std::pair<weight_type,std::string>>;

// type aliases for the solver component

// using intid = uint32_t;
using intid = weight_type;
template<typename Coeff>
using icpair = std::pair<intid,Coeff>;
template<typename Coeff>
using icpairs = std::vector<icpair<Coeff>>;
using eqdata = std::vector<std::pair<intid,std::string>>;

}
#endif
