/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#include <fstream>
#include <stack>
#include <stdexcept>
#include <unistd.h> // getopt

#include "interface.h"
#include "parser.h"

namespace pyred {

float time_diff(std::clock_t start, std::clock_t end) {
  return (float(end-start)/CLOCKS_PER_SEC);
}

/**********
 * Logger *
 **********/

Logger::Logger(int verb, int lev): verbosity_val{verb}, level{lev} {}

void Logger::set_verbosity(int verb) {verbosity_val = verb;}

int Logger::verbosity() {return verbosity_val;}

Logger& Logger::operator()(int lev) {
  std::cout << std::flush;
  level = lev;
  return *this;
}

Logger& Logger::operator<<(std::ostream& (*omanip)(std::ostream&)) {
  if (level <= verbosity_val) std::cout << (*omanip);
  return *this;
}

/**********
 * Config *
 **********/

Logger Config::log_val{1};
int Config::coeff_cls_val{1};
int Config::solver_type_val{1};
bool Config::backward_val{true};

void Config::set_coeff_cls(int ccls) {
  coeff_cls_val = ccls;
}

void Config::set_solver_type(int st) {
  solver_type_val = st;
}

void Config::set_backward(bool bkwd) {
  backward_val = bkwd;
}

void Config::set_verbosity(int verb) {
  log_val.set_verbosity(verb);
}

int Config::coeff_cls() {
  return coeff_cls_val;
}

int Config::solver_type() {
  return solver_type_val;
}

bool Config::backward() {
  return backward_val;
}

int Config::verbosity() {
  return log_val.verbosity();
}

Logger& Config::log(int lev) {
  return log_val(lev);
}

std::pair<std::string,std::string> Config::parse(const std::string& conf) {
  auto tmp = split(conf, ' ');
  tmp.insert(tmp.begin(), ""); // getopt() ignores the first element
  return parse(tmp);
}

std::pair<std::string,std::string> Config::parse(
    const std::vector<std::string>& conf) {
  // convert conf to char**
  int argc = conf.size();
  std::vector<char*> argv;
  argv.reserve(argc);
  for (const auto& opt: conf) {
    argv.push_back(const_cast<char*>(opt.c_str()));
  }
  int opt;
  int coeff_n{1};
  std::vector<std::string> vars{};
  std::string set1{};
  std::string infile{};
  std::string outfile{};
  optind = 1;
  while ((opt = getopt(argc, argv.data(), "1:c:i:o:s:b:t:v:")) != -1) {
    /*
     * -1:string symbol to set to one
     * -c:int use coefficient array of this size
     * -i:string declare symbol (comma separated or multiple -i options)
     * -o:string output file name
     * -s:int integer seed for random number generator
     * -b:int 0/1 to turn off/on backward insertion (default=1)
     * -t:int solver type
     */
    int seed;
    switch (opt) {
    case '1':
      set1 = std::string(optarg);
      break;
    case 'c':
      try {
        coeff_n = std::stoi(std::string(optarg));
      }
      catch(std::invalid_argument& e) {
        std::stringstream ss;
        ss << "Option c argument must be integer, but is "
           << optarg << std::endl;
        throw init_error(ss.str());
      }
      set_coeff_cls(2);
      break;
    case 'i':
      for (const auto& var: pyred::split(std::string(optarg), ',')) {
        vars.push_back(var);
      }
      break;
    case 'o':
      outfile = optarg;
      break;
    case 's':
      try {
        seed = std::stoi(std::string(optarg));
      }
      catch(std::invalid_argument& e) {
        std::stringstream ss;
        ss << "Seed argument must be integer, but is "
           << optarg << std::endl;
        throw init_error(ss.str());
      }
      pyred::random_seed(seed);
      break;
    case 'b':
      if (std::string(optarg) == "0") {
        set_backward(false);
      }
      else if (std::string(optarg) == "1") {
        set_backward(true);
      }
      else {
        std::stringstream ss;
        ss << "Option -b must either be 0 or 1, but is "
           << optarg << std::endl;
        throw init_error(ss.str());
      }
      break;
    case 't':
      try {
        set_solver_type(std::stoi(std::string(optarg)));
      }
      catch(std::invalid_argument& e) {
        std::stringstream ss;
        ss << "Solver type argument must be integer, but is "
           << optarg << std::endl;
        throw init_error(ss.str());
      }
      break;
    case 'v':
      try {
        set_verbosity(std::stoi(std::string(optarg)));
      }
      catch(std::invalid_argument& e) {
        std::stringstream ss;
        ss << "Verbosity argument must be integer, but is "
           << optarg << std::endl;
        throw init_error(ss.str());
      }
      break;
    default:
      std::stringstream ss;
      ss << "Invalid argument " << opt << std::endl;
      throw init_error(ss.str());
    }
  }
  if (coeff_cls() == 1) {
    Coeff_int::init(vars, coeff_n, set1);
  }
  else {
    Coeff_vec::init(vars, coeff_n, set1);
  }
  if (optind == argc-1) {
    infile = argv[optind];
  }
  else if (optind < argc-1) {
    std::stringstream ss;
    ss << "Too many arguments" << std::endl;
    throw init_error(ss.str());
  }
  return {infile, outfile};
}

/**********
 * System *
 **********/

System::System(): sys{} {}
System::System(const std::vector<eqdata>& eqs): sys{eqs} {}
System::System(std::vector<eqdata>&& eqs): sys{std::move(eqs)} {}

System::System(const std::string& infile): sys{} {
  add(infile);
}

std::size_t System::size() const {
  return sys.size();
}

void System::reserve(const std::size_t sz) {
  sys.reserve(sz);
}

void System::add(const eqdata& eq) {
  sys.push_back(eq);
}

void System::add(eqdata&& eq) {
  sys.push_back(std::move(eq));
}

void System::add(const std::vector<eqdata>& eqs){
  reserve(sys.size() + eqs.size());
  for (const auto& eq: eqs) {
    sys.push_back(eq);
  }
}

void System::add(std::vector<eqdata>&& eqs){
  reserve(sys.size() + eqs.size());
  for (auto& eq: eqs) {
    sys.push_back(std::move(eq));
  }
}

void System::add_forward(const eqdata& eq, intid neq) {
  if (Config::coeff_cls() == 1) {
    add_forward_tmpl<Coeff_int>(eq, numsys_int, sols_int, neq);
  }
  else {
    add_forward_tmpl<Coeff_vec>(eq, numsys_vec, sols_vec, neq);
  }
}

void System::add(const std::string& infile) {
  Config::log(1) << "import equations" << std::flush;
  auto starttime = std::clock();
  std::string line;
  eqdata tmpeq;
  if (infile.empty()) {
    // read equations from stdin
    while (std::getline(std::cin, line)) {
      if (line.empty() && !tmpeq.empty()) {
        sys.push_back(std::move(tmpeq));
        tmpeq.clear();
      }
      else {
        tmpeq.push_back(split_intcoeff(line));
      }
    }
    if (!tmpeq.empty()) {
      sys.push_back(std::move(tmpeq));
    }
  }
  else {
    // import equations from file
    std::ifstream instream(infile);
    if (!instream.good()) {
      std::stringstream ss;
      ss << "failed reading file \"" << infile << "\"" << std::endl;
      throw input_error(ss.str());
    }
    while (std::getline(instream, line)) {
      if (line.empty() && !tmpeq.empty()) {
        sys.push_back(std::move(tmpeq));
        tmpeq.clear();
      }
      else {
        tmpeq.push_back(split_intcoeff(line));
      }
    }
  }
  if (!tmpeq.empty()) {
    sys.push_back(std::move(tmpeq));
  }
  auto endtime = std::clock();
  Config::log(1) << ":   " << time_diff(starttime, endtime) << std::endl;
}

std::vector<intid> System::solve() {
  std::vector<intid> indep_eqs;
  if (Config::coeff_cls() == 1) {
    indep_eqs = solve_tmpl<Coeff_int>();
  }
  else {
    indep_eqs = solve_tmpl<Coeff_vec>();
  }
  return indep_eqs;
}

std::vector<intid> System::backward() {
  std::vector<intid> indep_eqnums;
  if (Config::coeff_cls() == 1) {
    Cache<Coeff_int>::clear();
    sols_int.clear();
    indep_eqnums = backward_tmpl<Coeff_int>(numsys_int, 2);
  }
  else {
    Cache<Coeff_vec>::clear();
    sols_vec.clear();
    indep_eqnums = backward_tmpl<Coeff_vec>(numsys_vec, 2);
  }
  return indep_eqnums;
}

const content_type& System::get_content() const {
  if (content.empty()) {
    throw init_error(
      "Content map is empty. System must be solved with backward insertion.");
  }
  return content;
}

std::pair<std::vector<intid>,std::vector<intid>> System::select(
  const std::vector<intid>& mandatory, const std::vector<intid>& if_reduced) {
  /*
   * 'mandatory':
   * - Vector of integrals for which a reduction formula is required.
   * 'if_reduced': V
   * - Vector of integrals to reduce, if there is a reduction formula known.
   * return: pair(neededeqs, unreduced)
   * 'neededeqs'
   * - The numbers of all equations which are needed to reduce integrals
   *   from 'mandatory' and 'if_reduced'.
   * 'unreduced':
   * - All integrals which appear in reduction formulas of mandatory integrals.
   * - All mandatory integrals for which there is no reduction formula known.
   */
  // Stack of untreated integrals for which the dependencies must be looked up.
  get_content(); // Throw error if 'content' is empty.
  std::stack<intid,std::vector<intid>> intstack;
  std::unordered_set<intid> neededset;
  std::unordered_set<intid> mandatoryset;
  // Set of integrals which are regarded as master integrals.
  std::unordered_set<intid> unreducedset;
  for (const auto iid: mandatory) neededset.insert(iid);
  mandatoryset = neededset;
  for (const auto iid: if_reduced) neededset.insert(iid);
  for (const auto iid: neededset) intstack.push(iid);
  while (!intstack.empty()) {
    auto next = intstack.top();
    intstack.pop();
    const auto cntit = content.find(next);
    if (cntit != content.cend()) {
      // A reduction formula is known.
      for (const auto& dep: std::get<2>(cntit->second)) {
        // For each integral on which 'next' depends:
        // add it to the set of needed integrals.
        // If it was not already in the set, push it onto the stack.
        auto isnew = neededset.insert(dep).second;
        if (isnew) intstack.push(dep);
      }
      if (mandatoryset.find(next) != mandatoryset.cend()) {
        // If 'next' is in 'mandatory', treat integral content masters.
        for (const auto& iid: std::get<1>(cntit->second)) {
          unreducedset.insert(iid);
        }
      }
    }
    else {
      // No reduction formula is known.
      // Dependencies always correspond to (inserted) equations,
      // i.e. this branch is only reached for unreduced integrals
      // in 'mandatory' and 'if_reduced'.
      if (mandatoryset.find(next) != mandatoryset.cend()) {
        // 'next' is in 'mandatory', but there is no reduction formula.
        // -> Regard it as a master integral.
        unreducedset.insert(next);
      }
      else {
        // 'next' is in 'if_reduced', but there is no reduction formula.
        // -> Remove it from neededset. It won't be added again,
        //    because it was never inserted anywhere.
        // Do not regard it as a master integral.
        neededset.erase(next);
      }
    }
  }
  // For each integral id in 'neededset':
  // push the corresponding equation number on 'neededeqs',
  // if a reduction formula exists.
  std::vector<intid> neededeqs;
  neededeqs.reserve(neededset.size());
  for (const auto iid: neededset) {
    const auto cntit = content.find(iid);
    if (cntit != content.cend()) {
      neededeqs.push_back(std::get<0>(cntit->second));
    }
  }
  std::sort(neededeqs.begin(), neededeqs.end());
  // Convert 'unreducedset' to a vector.
  std::vector<intid> unreduced;
  for (const auto iid: unreducedset) unreduced.push_back(iid);
  std::sort(unreduced.begin(), unreduced.end());
  return {neededeqs, unreduced};
}

std::vector<eqdata> System::retrieve() {
  return std::move(sys);
}

std::vector<eqdata> System::retrieve(const std::vector<intid>& eqnums) {
  auto eqns = eqnums;
  std::sort(eqns.begin(), eqns.end());
  if (!eqns.empty() && eqns.back() >= eqns.size()) {
    throw input_error("System::retrieve() argument out of range");
  }
  std::vector<eqdata> tmpsys;
  tmpsys.reserve(eqns.size());
  intid prev = eqns.size();
  for (auto eqn: eqns) {
    if (eqn != prev) {
      tmpsys.push_back(std::move(sys[eqn]));
      prev = eqn;
    }
  }
  sys.clear();
  content.clear();
  return tmpsys;
}

}
